/* libiso8601/src/libiso8601/calc.c
 *
 *  (c)2006, Laurence Withers, <l@lwithers.me.uk>.
 *  Released under the GNU GPLv2. See file COPYING or
 *  http://www.gnu.org/copyleft/gpl.html for details.
*/



/* CALCULATION ROUTINES
 *
 *  This file contains calculation routines used internally in the library. Our date format is the
 *  number of days elapsed since 0000-001 (so that date would be 0, -0001-365 would be -1, etc.).
 *  Time is represented as the number of seconds elapsed since midnight at the start of the day
 *  (so 0 is 00:00:00 and 86400 is 24:00:00, which we take to be infinitesimally before the start
 *  of the next day).
 */



int iso8601_isleap(int year)
{
        if(year % 4) return 0;
        if(year % 100) return 1;
        if(!(year % 400)) return 1;
        return 0;
}



/* struct monthcount, _days_in_month_common[], _days_in_month_leap[]
 *
 *  Tables of the number of days in each month, and the number of days elapsed since the start of
 *  the year for each month.
 */
struct monthcount {
    int elapsed, days;
};

static const struct monthcount _days_in_month_common[] = {
    { 0, 31 },
    { 31, 28 },
    { 59, 31 },
    { 90, 30 },
    { 120, 31 },
    { 151, 30 },
    { 181, 31 },
    { 212, 31 },
    { 243, 30 },
    { 273, 31 },
    { 304, 30 },
    { 334, 31 }
};

static const struct monthcount _days_in_month_leap[] = {
    { 0, 31 },
    { 31, 29 },
    { 60, 31 },
    { 91, 30 },
    { 121, 31 },
    { 152, 30 },
    { 182, 31 },
    { 213, 31 },
    { 244, 30 },
    { 274, 31 },
    { 305, 30 },
    { 335, 31 }
};



void iso8601_to_cal(int* year, int* month, int* day, const struct iso8601_date* date)
{
    div_t qr;
    const struct monthcount* mc;
    int ndays = date->day;

    // Each 400 years have 97 leap days, giving 365*400+97 = 146097 days in 400 years
    qr = div(ndays, 146097);
    *year = qr.quot * 400;
    ndays = qr.rem;

    // ensure that we always end up with between 0 and 146096 days remaining
    if(ndays < 0) {
        ndays += 146097;
        *year -= 400;
    }

    // we insert `fake' leap days for years 101, 201, 301
    if(ndays >= 36890) ++ndays;
    if(ndays >= 73415) ++ndays;
    if(ndays >= 109940) ++ndays;

    // each remaining 100 year block has 24 leap days, giving 365*100+24 = 36524 days
    qr = div(ndays, 36525);
    *year += qr.quot * 100;
    ndays = qr.rem;

    // each 4-year block has 1 leap day, giving 365*4 + 1 = 1461 days
    qr = div(ndays, 1461);
    *year += qr.quot * 4;
    ndays = qr.rem;

    // the first year of a 4-year block has 1 leap day, 366 days
    if(ndays >= 366) {
        --ndays; // pretend to have dealt with leap day

        // 365 days per remaining year
        qr = div(ndays, 365);
        *year += qr.quot;
        ndays = qr.rem;
    }

    // now we simply have number of days elapsed since day 001 in `year'.
    mc = iso8601_isleap(*year) ? _days_in_month_leap : _days_in_month_common;
    *month = 1;
    while(ndays >= mc->days) {
        ++*month;
        ndays -= mc->days;
        ++mc;
    }
    *day = ndays + 1;
}



int iso8601_from_cal(struct iso8601_date* date, int year, int month, int day)
{
    div_t qr;
    const struct monthcount* mc;

    // check for range / domain errors
    if(year < -5879609 || year > 5879609) {
        errno = ERANGE;
        return -1;
    }

    mc = iso8601_isleap(year) ? _days_in_month_leap : _days_in_month_common;
    if(month < 1 || month > 12 || day < 0 || day > mc[month - 1].days) {
        errno = EDOM;
        return -1;
    }

    // Each 400 years have 97 leap days, giving 365*400+97 = 146097 days in 400 years
    qr = div(year, 400);
    date->day = qr.quot * 146097;
    year = qr.rem;

    // ensure we have between 0 and 399 years
    if(year < 0) {
        date->day -= 146097;
        year += 400;
    }

    // excluding leap days, there are 365 days per year
    date->day += 365 * year;
    date->day += (year + 3) / 4; // there is one leap year for every four years
    date->day -= (year - 1) / 100; // but one less for every century over 0

    // now get number of days elapsed up to start of month
    date->day += mc[month - 1].elapsed;

    // and add number of days elapsed sinced start of month
    date->day += day - 1;

    return 0;
}



/* options for text editors
kate: replace-trailing-space-save true; space-indent true; tab-width 4;
vim: expandtab:ts=4:sw=4
*/
